// Part 1
// a)  Call-by-value version
/*def bob(x: Int): Int = { println("Bob"); x + 1 }

def joe(x: Int,y: Int): Int = {
    println("Joe");
    val a = x; val b = y; println(a+b);
    a+b
}
def ron(x: Int, y: Int, z: Int) {
    println("Ron")
    println(x + y)
    println(x + y + z)
}

ron(bob(joe(bob(1),2)),3,4)
*/
/*
Because all parameters are call-by-value, the value for each
parameter needs to be computed before the body of a function is executed. 
If the value passed as a parameter is the result of a function call, this
has to run first. For instance, when evaluating joe(bob(1),2), bob(1) 
is evaluated before joe. The output will therefore show "Bob" first, and then 
"Joe". 

Output: 

Bob - bob(1) is evaluated first so that its value can be passed to joe(bob(1),2)
Joe - printed after Bob returned its result. 
4   - a and b have been set to the value of the parameters. result of a+b. 
Bob - body of Bob is executed after joe(bob(1),2) has been evaluated. 
Ron - body of Ron is executed after bob(joe(bob(1),2)) has been evaluated. 
8   - part of the computation within ron
12  - part of the computation within ron.
*/

// b)  Call-by-name version
def bob(x: =>Int): Int = { println("Bob"); x + 1 }

def joe(x: =>Int,y: =>Int): Int = {
    println("Joe");
    val a = x; val b = y; println(a+b);
    a+b
}

def ron(x: =>Int, y: =>Int, z: =>Int) {
    println("Ron")
    println(x + y)
    println(x + y + z)
}

ron(bob(joe(bob(1),2)),3,4)

/*

With by-name parameters, if a function call is passed as a parameter, the call is evaluated
every time it is used within the body of the function. 

Output: 
Ron - first line of ron. Because all parameters are by-value and none have been accessed
      within ron yet, the output "Ron" is printed first.
Bob - second line of ron accesses x and y. x is the function call bob(joe(bob(1),2)), 
      which is evaluated at this time. "Bob" is the first statement within the body of bob. 
Joe - second statement within joe references by-value parameter x, which is joe(bob(1),2). 
      The print statement is the first line of joe. 
Bob - After "Joe" has been printed in joe, the val assignment accesses x, which is bob(1).
4   - output of println(a+b) within joe. 
      Once the value of evaluating x (i.e. bob(1) ) has been assigned to a, it is not evaluated again. 
8   - part of the main call println(x+y). joe(bob(1),2) has been evaluated and returned the result 5.
      y is 3. 
Bob - The next line of the main call requests x again. So bob(joe(bob(1),2)) is evaluated a second time.
Joe - Same as above. 
Bob - Same as above. 
4   - Same as above
12  - computation for printn(x + y + z) finishes.
*/

// b)
//Call-by-value version

/*def buggy(x: Int): Int = {
     println(x);
     buggy(x - 1)
}
def foo(x: Int, y: Int): Int = {
    println(x)
    x + 2
}
println(foo(1,buggy(10))) 
*/

/* 
Output: 
Program won't terminate. All parameters are call-by-value so buggy(10) needs to be evaluated
before foo(1,buggy(10)). Buggy contains a recursive call without terminating condition, so the program
gets stuck in an infinite loop.
*/

//Call-by-name version
def buggy(x: =>Int): Int = {
     println(x);
     buggy(x - 1)
}
def foo(x: =>Int, y: =>Int): Int = {
    println(x)
    x + 2
}
println(foo(1,buggy(10))) 

/* 
Output: 
1 - foo starts running and print out the x parameter
3 - foo returns x+2, the result is printed
buggy(10) is never evaluated because the call-by-name parameter is never used in 
the body of foo.
*/

